__all__ = (
    "TypeLensError",
    "TypeViewError",
    "ParameterViewError",
)


class TypeLensError(Exception):
    """Base class for library exceptions."""


class TypeViewError(TypeLensError):
    """Base class for TypeView exceptions."""


class ParameterViewError(TypeLensError):
    """Base class for ParameterView exceptions."""



# Configuration file for the Sphinx documentation builder.
from __future__ import annotations

import os
from datetime import datetime
from functools import partial
from typing import TYPE_CHECKING, Any

from type_lens.__metadata__ import __project__, __version__

if TYPE_CHECKING:
    from sphinx.addnodes import document
    from sphinx.application import Sphinx

__all__ = (
    "delayed_setup",
    "setup",
    "update_html_context",
)

# -- Environmental Data ------------------------------------------------------
suppress_warnings = ["config.cache"]

# -- Project information -----------------------------------------------------
current_year = datetime.now().year
project = __project__
copyright = f"{current_year}, Litestar Organization"
release = os.getenv("_TYPE_LENS_DOCS_BUILD_VERSION", __version__.rsplit(".")[0])

# -- General configuration ---------------------------------------------------
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.napoleon",
    "sphinx.ext.autosectionlabel",
    "sphinx.ext.githubpages",
    "sphinx.ext.viewcode",
    "sphinx.ext.intersphinx",
    "sphinx_copybutton",
    "sphinx.ext.todo",
    "sphinx.ext.viewcode",
    "sphinx_click",
    "sphinx_toolbox.collapse",
    "sphinx_design",
    "sphinx_togglebutton",
    "sphinx_paramlinks",
]

intersphinx_mapping = {
    "python": ("https://docs.python.org/3", None),
}
PY_CLASS = "py:class"
PY_RE = r"py:.*"
PY_METH = "py:meth"
PY_ATTR = "py:attr"
PY_OBJ = "py:obj"

nitpicky = True
nitpick_ignore = []
nitpick_ignore_regex = []

napoleon_google_docstring = True
napoleon_include_special_with_doc = True
napoleon_use_admonition_for_examples = True
napoleon_use_admonition_for_notes = True
napoleon_use_admonition_for_references = False
napoleon_attr_annotations = True

autoclass_content = "class"
autodoc_class_signature = "separated"
autodoc_default_options = {"special-members": "__init__", "show-inheritance": True, "members": True}
autodoc_member_order = "bysource"
autodoc_typehints_format = "short"
autodoc_type_aliases = {"FilterTypes": "FilterTypes"}

autosectionlabel_prefix_document = True
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]

auto_pytabs_min_version = (3, 9)
auto_pytabs_max_version = (3, 12)
auto_pytabs_compat_mode = True

# -- Style configuration -----------------------------------------------------
html_theme = "litestar_sphinx_theme"
html_title = "Litestar Framework"
pygments_style = "lightbulb"
todo_include_todos = True

html_static_path = ["_static"]
templates_path = ["_templates"]
html_js_files = ["versioning.js"]
html_css_files = ["style.css"]

html_show_sourcelink = True
html_copy_source = True

html_context = {
    "source_type": "github",
    "source_user": "litestar-org",
    "source_repo": "type-lens",
    "current_version": "latest",
    "versions": [
        ("latest", "/latest"),
        ("development", "/main"),
    ],
    "version": release,
}

html_theme_options = {
    "logo_target": "/",
    "github_repo_name": "type-lens",
    "navigation_with_keys": True,
    "nav_links": [  # TODO(provinzkraut): I need a guide on extra_navbar_items and its magic :P
        {"title": "Home", "url": "index"},
        {
            "title": "Community",
            "children": [
                {
                    "title": "Contributing",
                    "summary": "Learn how to contribute to the Type Lens project",
                    "url": "contribution-guide",
                    "icon": "contributing",
                },
                {
                    "title": "Code of Conduct",
                    "summary": "Review the etiquette for interacting with the Litestar community",
                    "url": "https://github.com/litestar-org/.github?tab=coc-ov-file",
                    "icon": "coc",
                },
                {
                    "title": "Security",
                    "summary": "Overview of Litestar's security protocols",
                    "url": "https://github.com/litestar-org/.github?tab=coc-ov-file#security-ov-file",
                    "icon": "coc",
                },
            ],
        },
        {
            "title": "About",
            "children": [
                {
                    "title": "Litestar Organization",
                    "summary": "Details about the Litestar organization",
                    "url": "https://litestar.dev/about/organization",
                    "icon": "org",
                },
                {
                    "title": "Releases",
                    "summary": "Explore the release process, versioning, and deprecation policy for Litestar",
                    "url": "releases",
                    "icon": "releases",
                },
            ],
        },
        {
            "title": "Release notes",
            "children": [
                {
                    "title": "1.x Changelog",
                    "url": "changelog",
                    "summary": "All changes in the 1.x series",
                },
            ],
        },
        {
            "title": "Help",
            "children": [
                {
                    "title": "Discord Help Forum",
                    "summary": "Dedicated Discord help forum",
                    "url": "https://discord.gg/litestar",
                    "icon": "coc",
                },
                {
                    "title": "GitHub Discussions",
                    "summary": "GitHub Discussions",
                    "url": "https://github.com/litestar-org/type-lens/discussions",
                    "icon": "coc",
                },
                {
                    "title": "Stack Overflow",
                    "summary": "We monitor the <code><b>litestar</b></code> tag on Stack Overflow",
                    "url": "https://stackoverflow.com/questions/tagged/litestar",
                    "icon": "coc",
                },
            ],
        },
        {"title": "Sponsor", "url": "https://github.com/sponsors/Litestar-Org", "icon": "heart"},
    ],
}


def update_html_context(
    app: Sphinx, pagename: str, templatename: str, context: dict[str, Any], doctree: document
) -> None:
    context["generate_toctree_html"] = partial(context["generate_toctree_html"], startdepth=0)


def delayed_setup(app: Sphinx) -> None:
    """
    When running linkcheck Shibuya causes a build failure, and checking
    the builder in the initial `setup` function call is not possible, so the check
    and extension setup has to be delayed until the builder is initialized.
    """
    if app.builder.name == "linkcheck":
        return

    app.setup_extension("shibuya")
    # app.connect("html-page-context", update_html_context)  # TODO(provinkraut): fix


def setup(app: Sphinx) -> dict[str, bool]:
    app.connect("builder-inited", delayed_setup, priority=0)
    app.setup_extension("litestar_sphinx_theme")
    return {"parallel_read_safe": True, "parallel_write_safe": True}



from __future__ import annotations

import sys
from dataclasses import dataclass
from typing import Any, List, Optional, Type, Union, cast

import pytest
from typing_extensions import Annotated

from type_lens import CallableView, ParameterView, TypeView


def test_invalid() -> None:
    class Foo: ...

    # Avoid linting deficiencies across versions.
    instance = cast(Type[Foo], Foo())

    with pytest.raises(ValueError) as e:
        CallableView.from_callable(instance)

    error = str(e.value)
    assert error.startswith("<tests.test_callable_view.test_invalid.<locals>.Foo object")
    assert error.endswith("is not a valid callable.")


def test_no_return_annotation() -> None:
    def fn():  # type: ignore[no-untyped-def]
        return

    function_view = CallableView.from_callable(fn)
    assert function_view.parameters == ()
    assert function_view.return_type == TypeView(None)


def test_return_annotation() -> None:
    def fn() -> int:
        return 4

    function_view = CallableView.from_callable(fn)
    assert function_view.parameters == ()
    assert function_view.return_type == TypeView(int)


def test_untyped_param() -> None:
    def fn(foo):  # type: ignore[no-untyped-def]
        return foo

    function_view = CallableView.from_callable(fn)
    assert function_view.parameters == (ParameterView("foo", TypeView(Any)),)


def test_typed_param() -> None:
    def fn(foo: int) -> int:
        return foo

    function_view = CallableView.from_callable(fn)
    assert function_view.parameters == (ParameterView("foo", TypeView(int)),)


def test_fix_annotated_optional_type_hints() -> None:
    def fn1(foo: Annotated[Optional[int], "d"] = None) -> Optional[int]:  # noqa: UP007
        return foo

    def fn2(foo: Annotated[Optional[int], "d"] = None) -> Optional[int]:  # noqa: UP007
        return foo

    fn_view1 = CallableView.from_callable(fn1, include_extras=True)
    assert fn_view1.parameters == (ParameterView("foo", TypeView(Annotated[Union[int, None], "d"]), default=None),)

    fn_view2 = CallableView.from_callable(fn2, include_extras=True)
    assert fn_view2.parameters == (ParameterView("foo", TypeView(Annotated[Union[int, None], "d"]), default=None),)


def test_class() -> None:
    @dataclass
    class Foo:
        a: int
        b: str = "asdf"

    fn_view1 = CallableView.from_callable(Foo, include_extras=True)
    assert fn_view1.parameters == (
        ParameterView("a", TypeView(int)),
        ParameterView("b", TypeView(str), default="asdf"),
    )


def test_class_instance() -> None:
    @dataclass
    class Foo:
        a: int
        b: str = "asdf"

        def __call__(self, c: bool) -> bool:
            return c

    foo = Foo(1, "qwer")

    fn_view1 = CallableView.from_callable(foo, include_extras=True)
    assert fn_view1.parameters == (ParameterView("c", TypeView(bool)),)


def test_instance_method() -> None:
    class Foo:
        def method(self, c: bool) -> bool:
            return c

    foo = Foo()

    fn_view1 = CallableView.from_callable(foo.method, include_extras=True)
    assert fn_view1.parameters == (ParameterView("c", TypeView(bool)),)


@pytest.mark.parametrize(
    ("hint",),
    [
        (Optional[str],),
        (Union[str, None],),
        (Union[str, int, None],),
        (Optional[Union[str, int]],),
        pytest.param(
            Union[str, int], marks=pytest.mark.xfail(sys.version_info < (3, 11), reason="Weird optional coercion")
        ),
        pytest.param(str, marks=pytest.mark.xfail(sys.version_info < (3, 11), reason="Weird optional coercion")),
    ],
)
def test_parameters_with_none_default(hint: Any) -> None:
    def fn(plain: hint = None, annotated: Annotated[hint, ...] = None) -> None: ...  # pyright: ignore

    fn_view = CallableView.from_callable(fn, localns=locals(), include_extras=True)
    plain_param, annotated_param = fn_view.parameters
    assert plain_param.type_view.annotation == annotated_param.type_view.annotation


def test_evaluated_runtime_type() -> None:
    @dataclass
    class Foo:
        a: list[int] | None = None

    fn_view1 = CallableView.from_callable(Foo, include_extras=True)

    if sys.version_info < (3, 10):
        assert fn_view1.parameters == (ParameterView("a", TypeView(Union[List[int], None]), default=None),)  # pyright: ignore
    else:
        assert fn_view1.parameters == (ParameterView("a", TypeView(Union[list[int], None]), default=None),)  # pyright: ignore



from __future__ import annotations

import typing as t
from collections import abc, defaultdict, deque

import typing_extensions as te

from type_lens.types.builtins import UNION_TYPES

__all__ = ("unwrap_annotation", "SAFE_GENERIC_ORIGIN_MAP", "INSTANTIABLE_TYPE_MAPPING")

SAFE_GENERIC_ORIGIN_MAP: te.Final[dict[object, object]] = {
    set: t.AbstractSet,
    defaultdict: t.DefaultDict,
    deque: t.Deque,
    dict: t.Dict,
    frozenset: t.FrozenSet,
    list: t.List,
    tuple: t.Tuple,
    abc.Mapping: t.Mapping,
    abc.MutableMapping: t.MutableMapping,
    abc.MutableSequence: t.MutableSequence,
    abc.MutableSet: t.MutableSet,
    abc.Sequence: t.Sequence,
    abc.Set: t.AbstractSet,
    abc.Collection: t.Collection,
    abc.Container: t.Container,
    abc.ItemsView: t.ItemsView,
    abc.KeysView: t.KeysView,
    abc.MappingView: t.MappingView,
    abc.ValuesView: t.ValuesView,
    abc.Iterable: t.Iterable,
    abc.Iterator: t.Iterator,
    abc.Generator: t.Generator,
    abc.Reversible: t.Reversible,
    abc.Coroutine: t.Coroutine,
    abc.AsyncGenerator: t.AsyncGenerator,
    abc.AsyncIterable: t.AsyncIterable,
    abc.AsyncIterator: t.AsyncIterator,
    abc.Awaitable: t.Awaitable,
    **{union_t: t.Union for union_t in UNION_TYPES},
}
"""A mapping of types to equivalent types that are safe to be used as generics across all Python versions.

This is necessary because occasionally we want to rebuild a generic outer type with different args, and types such as
``collections.abc.Mapping``, are not valid generic types in Python 3.8.
"""

_WRAPPER_TYPES: te.Final = {te.Annotated, te.Required, te.NotRequired}
"""Types that always contain a wrapped type annotation as their first arg."""

INSTANTIABLE_TYPE_MAPPING: te.Final = {
    t.AbstractSet: set,
    t.DefaultDict: defaultdict,
    t.Deque: deque,
    t.Dict: dict,
    t.FrozenSet: frozenset,
    t.List: list,
    t.Mapping: dict,
    t.MutableMapping: dict,
    t.MutableSequence: list,
    t.MutableSet: set,
    t.Sequence: list,
    t.Set: set,
    t.Tuple: tuple,
    abc.Mapping: dict,
    abc.MutableMapping: dict,
    abc.MutableSequence: list,
    abc.MutableSet: set,
    abc.Sequence: list,
    abc.Set: set,
    defaultdict: defaultdict,
    deque: deque,
    dict: dict,
    frozenset: frozenset,
    list: list,
    set: set,
    tuple: tuple,
}
"""A mapping of types to equivalent types that are safe to instantiate."""


def unwrap_annotation(annotation: t.Any) -> tuple[t.Any, tuple[t.Any, ...], set[t.Any]]:
    """Remove "wrapper" annotation types, such as ``Annotated``, ``Required``, and ``NotRequired``.

    Note:
        ``annotation`` should have been retrieved from :func:`get_type_hints()` with ``include_extras=True``. This
        ensures that any nested ``Annotated`` types are flattened according to the PEP 593 specification.

    Args:
        annotation: A type annotation.

    Returns:
        A tuple of the unwrapped annotation and any ``Annotated`` metadata, and a set of any wrapper types encountered.
    """
    origin = te.get_origin(annotation)
    wrappers = set()
    metadata = []
    while origin in _WRAPPER_TYPES:
        wrappers.add(origin)
        annotation, *meta = te.get_args(annotation)
        metadata.extend(meta)
        origin = te.get_origin(annotation)
    return annotation, tuple(metadata), wrappers



from __future__ import annotations

from inspect import Parameter
from typing import Any

from type_lens.parameter_view import ParameterView
from type_lens.type_view import TypeView
from type_lens.types.empty import Empty


def test_param_view() -> None:
    """Test ParameterView."""
    param = Parameter("foo", Parameter.POSITIONAL_OR_KEYWORD, annotation=int)
    param_view = ParameterView.from_parameter(param, {"foo": int})
    assert param_view.name == "foo"
    assert param_view.default is Empty
    assert param_view.type_view.annotation is int


def test_from_parameter_coerces_empty_annotation() -> None:
    param = Parameter("foo", Parameter.POSITIONAL_OR_KEYWORD)

    assert ParameterView.from_parameter(param, {}).type_view == TypeView(Any)


def test_param_view_has_default_predicate() -> None:
    """Test ParameterView.has_default."""
    param = Parameter("foo", Parameter.POSITIONAL_OR_KEYWORD, annotation=int)
    param_view = ParameterView.from_parameter(param, {"foo": int})
    assert param_view.has_default is False

    param = Parameter("foo", Parameter.POSITIONAL_OR_KEYWORD, annotation=int, default=42)
    param_view = ParameterView.from_parameter(param, {"foo": int})
    assert param_view.has_default is True


def test_param_view_repr() -> None:
    assert repr(ParameterView("str")).replace("typing.", "") == "ParameterView('str', TypeView(Any))"
    assert repr(ParameterView("str", TypeView(int))) == "ParameterView('str', TypeView(int))"
    assert (
        repr(ParameterView("str", default=True)).replace("typing.", "")
        == "ParameterView('str', TypeView(Any), default=True)"
    )



from __future__ import annotations

import sys
import types
import typing
from typing import Any

from type_lens.type_view import TypeView

__all__ = [
    "get_type_hints",
]


def get_type_hints(
    obj: Any,
    globalns: dict[str, Any] | None = None,
    localns: dict[str, Any] | None = None,
    include_extras: bool = False,
) -> dict[str, Any]:
    """Provide a `get_type_hints` implementation that is consistent across python versions."""
    result: dict[str, Any] = _get_type_hints(obj, globalns=globalns, localns=localns, include_extras=include_extras)  # type: ignore[no-untyped-call]
    if sys.version_info < (3, 11):  # pragma: no cover
        result = fix_annotated_optional_type_hints(result)

    return result


def fix_annotated_optional_type_hints(
    hints: dict[str, typing.Any],
) -> dict[str, typing.Any]:  # pragma: no cover
    """Normalize `Annotated` interacting with `get_type_hints` in versions <3.11.

    https://github.com/python/cpython/issues/90353.
    """
    for param_name, hint in hints.items():
        type_view = TypeView(hint)
        if type_view.is_union and type_view.inner_types[0].is_annotated:
            hints[param_name] = type_view.inner_types[0].raw
    return hints


if sys.version_info >= (3, 10):
    _get_type_hints = typing.get_type_hints  # pyright: ignore

else:
    from eval_type_backport import eval_type_backport

    @typing.no_type_check
    def _get_type_hints(  # noqa: C901
        obj: typing.Any,
        globalns: dict[str, typing.Any] | None = None,
        localns: dict[str, typing.Any] | None = None,
        include_extras: bool = False,
    ) -> dict[str, typing.Any]:  # pragma: no cover
        """Backport from python 3.10.8, with exceptions.

        * Use `_forward_ref` instead of `typing.ForwardRef` to handle the `is_class` argument.
        * `eval_type_backport` instead of `eval_type`, to backport syntax changes in Python 3.10.

        https://github.com/python/cpython/blob/aaaf5174241496afca7ce4d4584570190ff972fe/Lib/typing.py#L1773-L1875
        """
        if getattr(obj, "__no_type_check__", None):
            return {}
        # Classes require a special treatment.
        if isinstance(obj, type):
            hints = {}
            for base in reversed(obj.__mro__):
                if globalns is None:
                    base_globals = getattr(sys.modules.get(base.__module__, None), "__dict__", {})
                else:
                    base_globals = globalns
                ann = base.__dict__.get("__annotations__", {})
                if isinstance(ann, types.GetSetDescriptorType):
                    ann = {}
                base_locals = dict(vars(base)) if localns is None else localns
                if localns is None and globalns is None:
                    # This is surprising, but required.  Before Python 3.10,
                    # get_type_hints only evaluated the globalns of
                    # a class.  To maintain backwards compatibility, we reverse
                    # the globalns and localns order so that eval() looks into
                    # *base_globals* first rather than *base_locals*.
                    # This only affects ForwardRefs.
                    base_globals, base_locals = base_locals, base_globals
                for name, value in ann.items():
                    if value is None:
                        value = type(None)
                    if isinstance(value, str):
                        value = _forward_ref(value, is_argument=False, is_class=True)

                    value = eval_type_backport(value, base_globals, base_locals)
                    hints[name] = value
            if not include_extras and hasattr(typing, "_strip_annotations"):
                return {k: typing._strip_annotations(t) for k, t in hints.items()}
            return hints

        if globalns is None:
            if isinstance(obj, types.ModuleType):
                globalns = obj.__dict__
            else:
                nsobj = obj
                # Find globalns for the unwrapped object.
                while hasattr(nsobj, "__wrapped__"):
                    nsobj = nsobj.__wrapped__
                globalns = getattr(nsobj, "__globals__", {})
            if localns is None:
                localns = globalns
        elif localns is None:
            localns = globalns
        hints = getattr(obj, "__annotations__", None)
        if hints is None:
            # Return empty annotations for something that _could_ have them.
            if isinstance(obj, typing._allowed_types):
                return {}

            raise TypeError(f"{obj!r} is not a module, class, method, " "or function.")
        defaults = typing._get_defaults(obj)
        hints = dict(hints)
        for name, value in hints.items():
            if value is None:
                value = type(None)
            if isinstance(value, str):
                # class-level forward refs were handled above, this must be either
                # a module-level annotation or a function argument annotation

                value = _forward_ref(
                    value,
                    is_argument=not isinstance(obj, types.ModuleType),
                    is_class=False,
                )
            value = eval_type_backport(value, globalns, localns)
            if name in defaults and defaults[name] is None:
                value = typing.Optional[value]
            hints[name] = value

        if not include_extras and hasattr(typing, "_strip_annotations"):
            return {k: typing._strip_annotations(t) for k, t in hints.items()}
        return hints


def _forward_ref(
    arg: typing.Any,
    is_argument: bool = True,
    *,
    is_class: bool = False,
) -> typing.ForwardRef:
    return typing.ForwardRef(arg, is_argument)



from __future__ import annotations

import inspect
import types
from typing import TYPE_CHECKING, Any, Callable

from type_lens.parameter_view import ParameterView
from type_lens.type_view import TypeView
from type_lens.typing import get_type_hints

__all__ = ("CallableView",)


if TYPE_CHECKING:
    from typing_extensions import Self


class CallableView:
    def __init__(self, fn: Callable, type_hints: dict[str, type]):
        self.callable = fn
        self.signature = getattr(fn, "__signature__", None) or inspect.signature(fn)

        return_annotation = type_hints.pop("return", None)
        self.return_type = TypeView(return_annotation)

        self.parameters = tuple(
            ParameterView.from_parameter(param, type_hints) for param in self.signature.parameters.values()
        )

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, CallableView):
            return False

        return bool(
            self.callable == other.callable
            and self.parameters == other.parameters
            and self.return_type == other.return_type
        )

    def __repr__(self) -> str:
        cls_name = self.__class__.__name__

        return f"{cls_name}({self.callable.__name__})"

    @classmethod
    def from_callable(
        cls: type[Self],
        fn: Callable,
        *,
        globalns: dict[str, Any] | None = None,
        localns: dict[str, Any] | None = None,
        include_extras: bool = False,
    ) -> Self:
        hint_fn = fn
        if not isinstance(fn, (type, types.FunctionType)):
            callable_ = getattr(fn, "__func__", None) or getattr(fn, "__call__", None)  # noqa: B004
            if not callable_:
                raise ValueError(f"{fn} is not a valid callable.")

            hint_fn = callable_

        result = get_type_hints(hint_fn, globalns=globalns, localns=localns, include_extras=include_extras)
        return cls(fn, result)


def _fix_annotated_optional_type_hints(
    hints: dict[str, Any],
) -> dict[str, Any]:  # pragma: no cover
    """Normalize `Annotated` interacting with `get_type_hints` in versions <3.11.

    https://github.com/python/cpython/issues/90353.
    """
    for param_name, hint in hints.items():
        type_view = TypeView(hint)
        if type_view.is_union and type_view.inner_types[0].is_annotated:
            hints[param_name] = type_view.inner_types[0].raw
    return hints



# ruff: noqa: UP006
from __future__ import annotations

import sys
from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    ForwardRef,
    List,
    Literal,
    Optional,
    Sequence,
    Tuple,
    TypedDict,
    TypeVar,
    Union,
)

import pytest
from typing_extensions import Annotated, NotRequired, Required, get_type_hints
from typing_extensions import Literal as ExtensionsLiteral

from type_lens import TypeView
from type_lens.types.builtins import NoneType

if TYPE_CHECKING:
    from typing import Final


T = TypeVar("T")


def _check_parsed_type(type_lens: TypeView, expected: dict[str, Any]) -> None:
    __tracebackhide__ = True
    for key, expected_value in expected.items():
        lens_value = getattr(type_lens, key)
        if lens_value != expected_value:
            pytest.fail(f"Expected {key} to be {expected_value}, got {lens_value} instead. TypeLens: {type_lens}")


_type_lens_int: Final = TypeView(int)


class _TD(TypedDict):
    req_int: Required[int]
    req_list_int: Required[List[int]]
    not_req_int: NotRequired[int]
    not_req_list_int: NotRequired[List[int]]
    ann_req_int: Required[Annotated[int, "foo"]]
    ann_req_list_int: Required[Annotated[List[int], "foo"]]


_typed_dict_hints: Final = get_type_hints(_TD, include_extras=True)


@pytest.mark.parametrize(
    ("annotation", "expected"),
    [
        (
            int,
            {
                "raw": int,
                "annotation": int,
                "origin": None,
                "args": (),
                "metadata": (),
                "is_annotated": False,
                "is_required": False,
                "is_not_required": False,
                "inner_types": (),
            },
        ),
        (
            List[int],
            {
                "raw": List[int],
                "annotation": List[int],
                "origin": list,
                "args": (int,),
                "metadata": (),
                "is_annotated": False,
                "is_required": False,
                "is_not_required": False,
                "inner_types": (TypeView(int),),
            },
        ),
        (
            Annotated[int, "foo"],
            {
                "raw": Annotated[int, "foo"],
                "annotation": int,
                "origin": None,
                "args": (),
                "metadata": ("foo",),
                "is_annotated": True,
                "is_required": False,
                "inner_types": (),
            },
        ),
        (
            Annotated[List[int], "foo"],
            {
                "raw": Annotated[List[int], "foo"],
                "annotation": List[int],
                "origin": list,
                "args": (int,),
                "metadata": ("foo",),
                "is_annotated": True,
                "is_required": False,
                "is_not_required": False,
                "inner_types": (TypeView(int),),
            },
        ),
        (
            _typed_dict_hints["req_int"],
            {
                "raw": _typed_dict_hints["req_int"],
                "annotation": int,
                "origin": None,
                "args": (),
                "metadata": (),
                "is_annotated": False,
                "is_required": True,
                "is_not_required": False,
                "inner_types": (),
            },
        ),
        (
            _typed_dict_hints["req_list_int"],
            {
                "raw": _typed_dict_hints["req_list_int"],
                "annotation": List[int],
                "origin": list,
                "args": (int,),
                "metadata": (),
                "is_annotated": False,
                "is_required": True,
                "is_not_required": False,
                "inner_types": (TypeView(int),),
            },
        ),
        (
            _typed_dict_hints["not_req_int"],
            {
                "raw": _typed_dict_hints["not_req_int"],
                "annotation": int,
                "origin": None,
                "args": (),
                "metadata": (),
                "is_annotated": False,
                "is_required": False,
                "is_not_required": True,
                "inner_types": (),
            },
        ),
        (
            _typed_dict_hints["not_req_list_int"],
            {
                "raw": _typed_dict_hints["not_req_list_int"],
                "annotation": List[int],
                "origin": list,
                "args": (int,),
                "metadata": (),
                "is_annotated": False,
                "is_required": False,
                "is_not_required": True,
                "inner_types": (TypeView(int),),
            },
        ),
        (
            _typed_dict_hints["ann_req_int"],
            {
                "raw": _typed_dict_hints["ann_req_int"],
                "annotation": int,
                "origin": None,
                "args": (),
                "metadata": ("foo",),
                "is_annotated": True,
                "is_required": True,
                "is_not_required": False,
                "inner_types": (),
            },
        ),
        (
            _typed_dict_hints["ann_req_list_int"],
            {
                "raw": _typed_dict_hints["ann_req_list_int"],
                "annotation": List[int],
                "origin": list,
                "args": (int,),
                "metadata": ("foo",),
                "is_annotated": True,
                "is_required": True,
                "is_not_required": False,
                "inner_types": (TypeView(int),),
            },
        ),
    ],
)
def test_parsed_type_from_annotation(annotation: Any, expected: dict[str, Any]) -> None:
    """Test ParsedType.from_annotation."""
    _check_parsed_type(TypeView(annotation), expected)


def test_parsed_type_from_union_annotation() -> None:
    """Test ParsedType.from_annotation for Union."""
    annotation = Union[int, List[int]]
    expected = {
        "raw": annotation,
        "annotation": annotation,
        "origin": Union,
        "args": (int, List[int]),
        "metadata": (),
        "is_annotated": False,
        "is_required": False,
        "is_not_required": False,
        "inner_types": (TypeView(int), TypeView(List[int])),
    }
    _check_parsed_type(TypeView(annotation), expected)


@pytest.mark.parametrize("value", ["int", ForwardRef("int")])
def test_parsed_type_is_forward_ref_predicate(value: Any) -> None:
    """Test ParsedType with ForwardRef."""
    parsed_type = TypeView(value)
    assert parsed_type.is_forward_ref is True
    assert parsed_type.annotation == value
    assert parsed_type.origin is None
    assert parsed_type.args == ()
    assert parsed_type.metadata == ()
    assert parsed_type.is_annotated is False
    assert parsed_type.is_required is False
    assert parsed_type.is_not_required is False
    assert parsed_type.inner_types == ()


def test_parsed_type_is_type_var_predicate() -> None:
    """Test ParsedType.is_type_var."""
    assert TypeView(int).is_type_var is False
    assert TypeView(T).is_type_var is True
    assert TypeView(Union[int, T]).is_type_var is False  # pyright: ignore[reportGeneralTypeIssues]


def test_parsed_type_is_union_predicate() -> None:
    """Test ParsedType.is_union."""
    assert TypeView(int).is_union is False
    assert TypeView(Optional[int]).is_union is True
    assert TypeView(Union[int, None]).is_union is True
    assert TypeView(Union[int, str]).is_union is True


def test_parsed_type_is_optional_predicate() -> None:
    """Test ParsedType.is_optional."""
    assert TypeView(int).is_optional is False
    assert TypeView(Optional[int]).is_optional is True
    assert TypeView(Union[int, None]).is_optional is True
    assert TypeView(Union[int, None, str]).is_optional is True
    assert TypeView(Union[int, str]).is_optional is False


def test_parsed_type_is_subtype_of() -> None:
    """Test TypeView.is_subtype_of."""

    class Foo:
        pass

    class Bar(Foo):
        pass

    assert TypeView(bool).is_subtype_of(int) is True
    assert TypeView(bool).is_subtype_of(str) is False
    assert TypeView(Union[int, str]).is_subtype_of(int) is False
    assert TypeView(List[int]).is_subtype_of(list) is True
    assert TypeView(List[int]).is_subtype_of(int) is False
    assert TypeView(Optional[int]).is_subtype_of(int) is False
    assert TypeView(Union[bool, int]).is_subtype_of(int) is True

    assert TypeView(Foo).is_subtype_of(Foo) is True
    assert TypeView(Bar).is_subtype_of(Foo) is True
    assert TypeView(Union[bool, int]).is_subtype_of(int) is True


def test_is_subclass_of() -> None:
    class Foo:
        pass

    class Bar(Foo):
        pass

    assert TypeView(bool).is_subclass_of(int) is True
    assert TypeView(bool).is_subclass_of(str) is False
    assert TypeView(Union[int, str]).is_subclass_of(int) is False
    assert TypeView(List[int]).is_subclass_of(int) is False
    assert TypeView(Optional[int]).is_subclass_of(int) is False
    assert TypeView(None).is_subclass_of(int) is False
    assert TypeView(Literal[1]).is_subclass_of(int) is False
    assert TypeView(Union[float, str]).is_subclass_of(int) is False

    assert TypeView(bool).is_subclass_of(bool) is True
    assert TypeView(List[int]).is_subclass_of(list) is True
    assert TypeView(Foo).is_subclass_of(Foo) is True
    assert TypeView(Bar).is_subclass_of(Foo) is True


def test_parsed_type_has_inner_subtype_of() -> None:
    """Test ParsedType.has_type_of."""
    assert TypeView(List[int]).has_inner_subtype_of(int) is True
    assert TypeView(List[int]).has_inner_subtype_of(str) is False
    assert TypeView(List[Union[int, str]]).has_inner_subtype_of(int) is False


def test_parsed_type_equality() -> None:
    assert TypeView(int) == TypeView(int)
    assert TypeView(int) == TypeView(Annotated[int, "meta"])
    assert TypeView(int) != int
    assert TypeView(List[int]) == TypeView(List[int])
    assert TypeView(List[int]) != TypeView(List[str])
    assert TypeView(List[str]) != TypeView(Tuple[str])
    assert TypeView(Optional[str]) == TypeView(Union[str, None])


def test_tuple() -> None:
    assert TypeView(List[int]).is_tuple is False
    assert TypeView(List[int]).is_variadic_tuple is False

    assert TypeView(Tuple[int]).is_tuple is True
    assert TypeView(Tuple[int]).is_variadic_tuple is False

    assert TypeView(Tuple[int, int]).is_tuple is True
    assert TypeView(Tuple[int, int]).is_variadic_tuple is False

    assert TypeView(Tuple[int, ...]).is_tuple is True
    assert TypeView(Tuple[int, ...]).is_variadic_tuple is True

    assert TypeView(...).is_tuple is False
    assert TypeView(...).is_variadic_tuple is False


def test_strip_optional() -> None:
    # Non-optionals should return the original input
    assert TypeView(int).strip_optional() == TypeView(int)

    assert TypeView(Optional[int]).strip_optional() == TypeView(int)
    assert TypeView(Optional[Union[str, int]]).strip_optional() == TypeView(Union[str, int])
    assert TypeView(Union[str, int, None]).strip_optional() == TypeView(Union[str, int])


def test_repr() -> None:
    assert repr(TypeView(int)) == "TypeView(int)"
    assert repr(TypeView(Optional[str])) == "TypeView(Union[str, NoneType])"
    assert repr(TypeView(Literal["1", 2, True])) == "TypeView(Literal['1', 2, True])"
    assert repr(TypeView(Tuple[Literal["1"], int])) == "TypeView(Tuple[Literal['1'], int])"
    assert repr(TypeView(ExtensionsLiteral["1", 2, True])) == "TypeView(Literal['1', 2, True])"
    assert repr(TypeView(Any)) == "TypeView(Any)"


def test_is_none_type() -> None:
    assert TypeView(int).is_none_type is False
    assert TypeView(None).is_none_type is True
    assert TypeView(NoneType).is_none_type is True
    assert TypeView(Annotated[None, 4]).is_none_type is True
    assert TypeView(Union[int, None]).inner_types[1].is_none_type is True


def test_literal() -> None:
    assert TypeView(int).is_literal is False
    assert TypeView(Literal[4]).is_literal is True
    assert TypeView(4).is_literal is False


def test_allows_none() -> None:
    assert TypeView(int).allows_none is False
    assert TypeView(Optional[int]).allows_none is True
    assert TypeView(None).allows_none is True


@pytest.mark.parametrize(
    ("annotation", "expected"),
    [
        (int, None),
        (Optional[int], Union),
        (Union[int, None], Union),
        pytest.param(
            "int | str", Union, marks=pytest.mark.skipif(sys.version_info < (3, 10), reason="Requires Python 3.10")
        ),
        pytest.param(
            "list[int]", List, marks=pytest.mark.skipif(sys.version_info < (3, 9), reason="Requires Python 3.9")
        ),
    ],
)
def test_safe_generic_origin(annotation: Any, expected: Any) -> None:
    if isinstance(annotation, str):
        annotation = eval(annotation)
    assert TypeView(annotation).safe_generic_origin is expected


def test_repr_type() -> None:
    assert TypeView(int).repr_type == "int"
    assert TypeView(str).repr_type == "str"
    assert TypeView(Optional[int]).repr_type in ("Optional[int]", "Union[int, NoneType]", "Union[int, None]")
    assert TypeView(Literal[1, "two"]).repr_type == "Literal[1, 'two']"
    assert TypeView(Union[Literal[1, "two"], bool]).repr_type == "Union[Literal[1, 'two'], bool]"

    if sys.version_info >= (3, 9):
        assert TypeView(set[bool]).repr_type == "set[bool]"


def test_instantiatable_origin() -> None:
    assert TypeView(int).instantiable_origin == int
    assert TypeView(list).instantiable_origin == list
    assert TypeView(List[int]).instantiable_origin == list
    assert TypeView(Dict[int, int]).instantiable_origin == dict
    assert TypeView(Sequence[int]).instantiable_origin == list
    assert TypeView(TypeView).instantiable_origin == TypeView


def test_fallback_origin() -> None:
    assert TypeView(int).fallback_origin == int
    assert TypeView(list).fallback_origin == list
    assert TypeView(List).fallback_origin == list
    assert TypeView(List[str]).fallback_origin == list
    assert TypeView(Literal[1]).fallback_origin == Literal
    assert TypeView(Literal).fallback_origin == Literal

    if sys.version_info >= (3, 9):
        assert TypeView(set[bool]).fallback_origin == set



from __future__ import annotations

from inspect import Signature
from typing import TYPE_CHECKING, Any, Final

from type_lens.type_view import TypeView
from type_lens.types.empty import Empty, EmptyType

__all__ = ("ParameterView",)

if TYPE_CHECKING:
    from inspect import Parameter

    from typing_extensions import Self

_any_type_view = TypeView(Any)


class ParameterView:
    """Represents the parameters of a callable."""

    __slots__ = {
        "name": "The name of the parameter.",
        "default": "The default value of the parameter.",
        "type_view": "View of the parameter's annotation type.",
    }

    def __init__(self, name: str, type_view: TypeView = _any_type_view, *, default: Any | EmptyType = Empty) -> None:
        self.name: Final = name
        self.type_view: Final = type_view
        self.default: Final = default

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, ParameterView):
            return False

        return bool(self.name == other.name and self.type_view == other.type_view and self.default == other.default)

    def __repr__(self) -> str:
        cls_name = self.__class__.__name__

        args = [
            repr(self.name),
            repr(self.type_view) if self.type_view else None,
            f"default={self.default}" if self.default is not Empty else None,
        ]
        args_str = ", ".join(a for a in args if a is not None)
        return f"{cls_name}({args_str})"

    @property
    def has_default(self) -> bool:
        """Whether the parameter has a default value or not."""
        return self.default is not Empty

    @classmethod
    def from_parameter(cls, parameter: Parameter, fn_type_hints: dict[str, Any]) -> Self:
        """Initialize ParsedSignatureParameter.

        Args:
            parameter: inspect.Parameter
            fn_type_hints: mapping of names to types. Should be result of ``get_type_hints()``.

        Returns:
            ParsedSignatureParameter.
        """
        annotation = fn_type_hints.get(parameter.name, Any)

        return cls(
            name=parameter.name,
            default=Empty if parameter.default is Signature.empty else parameter.default,
            type_view=TypeView(annotation),
        )



from __future__ import annotations

import sys
from typing import Union

__all__ = ["UNION_TYPES", "NoneType"]

if sys.version_info >= (3, 10):
    from types import UnionType

    UNION_TYPES = {UnionType, Union}
else:  # pragma: no cover
    UNION_TYPES = {Union}

NoneType: type[None] = type(None)



from __future__ import annotations

from collections import abc
from collections.abc import Collection, Mapping
from typing import Any, AnyStr, Final, ForwardRef, Generic, Literal, TypeVar, Union, _SpecialForm

from typing_extensions import Annotated, NotRequired, Required, get_args, get_origin
from typing_extensions import Literal as ExtensionsLiteral

from type_lens.types.builtins import UNION_TYPES, NoneType
from type_lens.utils import INSTANTIABLE_TYPE_MAPPING, SAFE_GENERIC_ORIGIN_MAP, unwrap_annotation

__all__ = ("TypeView",)


T = TypeVar("T")


class TypeView(Generic[T]):
    """Represents a type annotation."""

    __slots__ = {
        "annotation": "The annotation with any 'wrapper' types removed, e.g. Annotated.",
        "args": "The result of calling get_args(annotation) after unwrapping Annotated, e.g. (int,).",
        "inner_types": "The type's generic args parsed as ParsedType, if applicable.",
        "metadata": "Any metadata associated with the annotation via Annotated.",
        "origin": "The result of calling get_origin(annotation) after unwrapping Annotated, e.g. list.",
        "fallback_origin": "The unsubscripted version of a type, distinct from 'origin' in that for non-generics, this is the original type.",
        "raw": "The annotation exactly as received.",
        "_wrappers": "A set of wrapper types that were removed from the annotation.",
    }

    def __init__(self, annotation: T) -> None:
        """Initialize ParsedType.

        Args:
            annotation: The type annotation. This should be extracted from the return of
                ``get_type_hints(..., include_extras=True)`` so that forward references are resolved and recursive
                ``Annotated`` types are flattened.

        Returns:
            ParsedType
        """
        unwrapped, metadata, wrappers = unwrap_annotation(annotation)
        origin = get_origin(unwrapped)

        args: tuple[Any, ...] = () if origin is abc.Callable else get_args(unwrapped)

        self.raw: Final[T] = annotation
        self.annotation: Final = unwrapped
        self.origin: Final = origin
        self.fallback_origin: Final = origin or unwrapped
        self.args: Final[tuple[Any, ...]] = args
        self.metadata: Final = metadata
        self._wrappers: Final = wrappers
        self.inner_types: Final = tuple(TypeView(arg) for arg in args)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, TypeView):
            return False

        if self.origin:
            self_origin = Union if self.is_union else self.origin
            other_origin = Union if other.is_union else other.origin
            return bool(self_origin == other_origin and self.inner_types == other.inner_types)

        return bool(self.annotation == other.annotation)

    def __repr__(self) -> str:
        cls_name = self.__class__.__name__
        return f"{cls_name}({self.repr_type})"

    @property
    def repr_type(self) -> str:
        """Returns a consistent, canonical repr of the contained annotation.

        Removes preceding `typing.` prefix for built-in typing constructs. Python's
        native repr for `typing` types is inconsistent across python versions!
        """
        # Literal/Union both appear to have no name on some versions of python.
        if self.is_literal:
            name = "Literal"
        elif self.is_union:
            name = "Union"
        elif isinstance(self.annotation, (type, _SpecialForm)) or self.origin:
            try:
                name = self.annotation.__name__  # pyright: ignore[reportAttributeAccessIssue]
            except AttributeError:
                # Certain _SpecialForm items have no __name__ python 3.8.
                name = self.annotation._name  # pyright: ignore[reportAttributeAccessIssue]
        else:
            name = repr(self.annotation)

        if self.origin:
            inner_types = ", ".join(t.repr_type for t in self.inner_types)
            name = f"{name}[{inner_types}]"

        return str(name)

    @property
    def allows_none(self) -> bool:
        """Whether the annotation supports being assigned ``None``."""
        return self.is_optional or self.is_none_type

    @property
    def instantiable_origin(self) -> Any:
        """An instantiable type that is consistent with the origin type of the annotation.

        Examples:
            >>> from type_lens import TypeView
            >>> from collections.abc import Sequence
            >>> TypeView(Sequence[int]).instantiable_origin
            <class 'list'>

        Returns:
            An instantiable type that is consistent with the origin type of the annotation.
        """
        return INSTANTIABLE_TYPE_MAPPING.get(self.fallback_origin, self.fallback_origin)

    @property
    def is_annotated(self) -> bool:
        """Whether the annotation was wrapped in Annotated or not.

        This would indicate that the annotation has metadata associated with it.
        """
        return Annotated in self._wrappers

    @property
    def is_collection(self) -> bool:
        """Whether the annotation is a collection type or not."""
        return self.is_subtype_of(Collection)

    @property
    def is_forward_ref(self) -> bool:
        """Whether the annotation is a forward reference or not."""
        return isinstance(self.annotation, (str, ForwardRef))

    @property
    def is_literal(self) -> bool:
        """Whether the annotation is a literal value or not."""
        return self.origin in {Literal, ExtensionsLiteral}

    @property
    def is_mapping(self) -> bool:
        """Whether the annotation is a mapping or not."""
        return self.is_subtype_of(Mapping)

    @property
    def is_non_string_collection(self) -> bool:
        """Whether the annotation is a non-string collection type or not."""
        return self.is_collection and not self.is_subtype_of((str, bytes))

    @property
    def is_none_type(self) -> bool:
        """Whether the annotation is NoneType or not."""
        return self.annotation in {None, NoneType}

    @property
    def is_not_required(self) -> bool:
        """Whether the annotation was wrapped in NotRequired or not."""
        return NotRequired in self._wrappers

    @property
    def is_optional(self) -> bool:
        """Whether the annotation is Optional or not."""
        return bool(self.is_union and NoneType in self.args)

    @property
    def is_required(self) -> bool:
        """Whether the annotation was wrapped in Required or not."""
        return Required in self._wrappers

    @property
    def is_tuple(self) -> bool:
        """Whether the annotation is a ``tuple`` or not."""
        return self.is_subclass_of(tuple)

    @property
    def is_type_var(self) -> bool:
        """Whether the annotation is a TypeVar or not."""
        return isinstance(self.annotation, TypeVar)

    @property
    def is_union(self) -> bool:
        """Whether the annotation is a union type or not."""
        return self.origin in UNION_TYPES

    @property
    def is_variadic_tuple(self) -> bool:
        """Whether the annotation is a ``tuple`` **and** is of unbounded length.

        Tuples like `tuple[int, ...]` represent a list-like unbounded sequence
        of a single type T.
        """
        return self.is_tuple and len(self.args) == 2 and self.args[1] == ...

    @property
    def safe_generic_origin(self) -> Any:
        """A type, safe to be used as a generic type across all supported Python versions.

        Examples:
            >>> from type_lens import TypeView
            >>> TypeView(dict[str, int]).safe_generic_origin
            typing.Dict
        """
        return SAFE_GENERIC_ORIGIN_MAP.get(self.fallback_origin)

    def has_inner_subtype_of(self, typ: type[Any] | tuple[type[Any], ...]) -> bool:
        """Whether any generic args are a subclass of the given type.

        Args:
            typ: The type to check, or tuple of types. Passed as 2nd argument to ``issubclass()``.

        Returns:
            Whether any of the type's generic args are a subclass of the given type.
        """
        return any(t.is_subtype_of(typ) for t in self.inner_types)

    def is_subtype_of(self, typ: Any | tuple[Any, ...], /) -> bool:
        """Whether the annotation is a subtype of the given type.

        Where ``self.annotation`` is a union type, this method will return ``True`` when all members of the union are
        a subtype of ``cl``, otherwise, ``False``.

        Args:
            typ: The type to check, or tuple of types. Passed as 2nd argument to ``issubclass()``.

        Returns:
            Whether the annotation is a subtype of the given type(s).
        """
        if self.origin:
            if self.is_union:
                return all(t.is_subtype_of(typ) for t in self.inner_types)

            return issubclass(self.origin, typ)

        if self.annotation is AnyStr:
            return TypeView(Union[str, bytes]).is_subtype_of(typ)
        return self.annotation is not Any and not self.is_type_var and self.is_subclass_of(typ)

    def is_subclass_of(self, typ: Any | tuple[Any, ...], /) -> bool:
        """Whether the annotation is a subclass of the given type.

        Args:
            typ: The type to check, or tuple of types. Passed as 2nd argument to ``issubclass()``.

        Returns:
            Whether the annotation is a subclass of the given type(s).
        """
        return isinstance(self.fallback_origin, type) and issubclass(self.fallback_origin, typ)

    def strip_optional(self) -> TypeView:
        if not self.is_optional:
            return self

        if len(self.args) == 2:
            return self.inner_types[0]

        args = tuple(a for a in self.args if a is not NoneType)
        non_optional = Union[args]  # type: ignore[valid-type]
        return TypeView(non_optional)



"""Metadata for the Project."""

from __future__ import annotations

import importlib.metadata

__all__ = ["__version__", "__project__"]

__version__ = importlib.metadata.version("type-lens")
"""Version of the project."""
__project__ = importlib.metadata.metadata("type-lens")["Name"]
"""Name of the project."""



from __future__ import annotations

from .callable_view import CallableView
from .parameter_view import ParameterView
from .type_view import TypeView
from .types.empty import Empty, EmptyType

__all__ = (
    "CallableView",
    "ParameterView",
    "TypeView",
    "EmptyType",
    "Empty",
)
